/*-----------------------------------------------------------------------------
**      gif_com.c
**
**      Handle comments in GIF files
**
**      2018-11-27  start
**
**
**
*/
#include <stdio.h>
#include <string.h>
#include <stdint.h>



FILE * fpin;

int     global_error_flag;

char    sigbuf[6];
const char sig87[] = { 'G', 'I', 'F', '8', '7', 'a' };
const char sig89[] = { 'G', 'I', 'F', '8', '9', 'a' };
int     flag_89;

typedef struct {
    int16_t     width;
    int16_t     height;
    int8_t      bits;
    uint8_t     bg_color;
    int8_t      ratio;
} Screen_Descriptor_t;

typedef struct {
    uint8_t     bits;
    uint16_t    delay;
    uint8_t     transparent_ix;
} Control_Extension_t;

typedef struct {
    int16_t     x_pos;
    int16_t     y_pos;
    int16_t     width;
    int16_t     height;
    uint8_t     bits;
} Image_Header_t;


Screen_Descriptor_t screen_descriptor;
Control_Extension_t control_extension;
Image_Header_t      image_header;       /* können mehrere sein ! */

int     img_cnt;

/*
**  get one byte from file
*/
unsigned char
getbyte( void ) {
    int c;
    if ( (c = fgetc( fpin )) >= 0 ){
        return (unsigned char)c;
    }
    global_error_flag = -1;
    return 0;
}


/*
**  get a short int from file
*/
int16_t
getshort( void ) {
    int c0;
    int c1;
    if ( (c0 = fgetc( fpin )) < 0 )
        global_error_flag = -1;
    if ( (c1 = fgetc( fpin )) < 0 )
        global_error_flag = -1;
    c1 <<= 8;
    c1 += c0;
    return (int16_t)c1;
}






int
chk_signature( void ) {
    fread( sigbuf, sizeof(sigbuf), 1, fpin );
    if( strncmp( sigbuf, sig87, sizeof(sigbuf) ) == 0 ) {
        flag_89 = 0;
        printf( "File is GIF87a\n" );
        return 1;
    }
    if( strncmp( sigbuf, sig89, sizeof(sigbuf) ) == 0 ) {
        flag_89 = 0;
        printf( "File is GIF89a\n" );
        return 2;
    }
    printf( "File is not a GIF file\n" );
    return 0;
}


int
get_screen_descriptor( void ) {
    Screen_Descriptor_t * sd = &screen_descriptor;
    int val;
    int ret = 0;

    sd->width = getshort();
    sd->height = getshort();
    sd->bits = getbyte();
    sd->bg_color = getbyte();
    sd->ratio = getbyte();

    printf( "Screen data:\n" );
    printf(
        "  width:  %d\n"
        "  height: %d\n", sd->width, sd->height );
    if( sd->bits & 0x80 ) {
        val = 2 << (sd->bits & 7);
        ret = val * 3;
        printf( "  Color table size: %d\n", val );
    } else {
        printf( "  No global color table\n" );
    }
    printf( "  Bits per pixel: %d\n", (sd->bits & 7) + 1 );
    printf( "  Background color index: %u\n", sd->bg_color );
    val = sd->ratio != 0 ? sd->ratio + 15 / 64 : 1;
    printf( "  Aspect ratio: %d\n\n", val );

    return ret;
}


void
get_control_extension( void ) {
    Control_Extension_t * cex = &control_extension;
    int     len = getbyte();

    cex->bits = getbyte();
    cex->delay = getshort();
    cex->transparent_ix = getbyte();
    printf( "Graphics control extension:\n" );
    if( cex->bits & 0x01 ) {
        printf( "  Transparent color index: %d\n", cex->transparent_ix );
    }
    printf( "  User input: %d\n", (cex->bits & 2) == 2 ? 1 : 0 );
    printf( "  Diposal method: %d\n", (cex->bits & 0x1C) >> 2 );
    getbyte();
}


void
get_extensions( void ) {
    int type;
    int len1;

    type = getbyte();
    switch( type ) {
        case 0xF9:  /*Graphic control extension */
            get_control_extension();
            break;

        case 0xFE:  /* Comment extension */
            len1 = getbyte();
            if( len1 > 0 ) {
                printf( "Comment:\n" );
                while( len1-- ) {
                    printf( "%c", getbyte() );
                }
                printf( "\n" );
            }
            getbyte();
            break;

        case 0xFF:  /* Application extension */
            len1 = getbyte();
            if( len1 != 11 )
                global_error_flag = -1;
            printf( "Application Extension: " );
            while( len1>0 ) {
                while (len1--) {
                    printf( "%c", getbyte() );
                }
            }
            printf( "\n" );
            len1=getbyte();
            printf( " %02X", len1 );
            while( len1 ) {
                while( len1-- )
                    printf( " %02X", getbyte() );
                len1 = getbyte();
                printf( " %02X", len1 );
            }
            printf( "\n" );
            break;

        case 0x01:  /* Plain text extension */
        default:
            len1 = getbyte();
            global_error_flag = -1;
            printf( "Unknown extension type: %02X\n\n", type );
            while (len1>0) {
                while (len1--) {
                    getbyte();
                }
                len1=getbyte();
            }
            break;
    }
    printf( "pos: %lX\n", ftell( fpin ) ); /* debug */
}



int
get_image( void ) {
    Image_Header_t * ih = &image_header;
    int     lc_size = 0;
    int len;

    ih->x_pos = getshort();
    ih->y_pos = getshort();
    ih->width = getshort();
    ih->height = getshort();
    ih->bits = getbyte();

    printf( "Image header %d\n  position x: %d  y: %d\n", ++img_cnt, ih->x_pos, ih->y_pos );
    printf( "  size: %d x %d\n", ih->width, ih->height );
    if( ih->bits & 0x80 ) {
        lc_size = 2 << (ih->bits & 7);
        printf( "  Local color table size: %d\n", lc_size );
        printf( "  Bits per pixel: %d\n", (ih->bits & 7) + 1 );
        lc_size *= 3;
    }
    printf( "  Interlace flag: %d\n", (ih->bits & 0x40) == 0x40? 1: 0 );

    printf( "  Initial code size: %d\n", getbyte() ); /* not part of the header ! */
    printf( "\n" );
    if( lc_size )
        fseek( fpin, lc_size, SEEK_CUR ); /* skip local color table */
    printf( "pos: %lX\n", ftell( fpin ) ); /* debug */

    /* read the data blocks */
    len = getbyte();
    while( len ) {
        while( len-- )
            getbyte();
        len = getbyte();
    }

    return lc_size;
}


int
main( int argc, char * argv[] ) {
    int val;
    char block_code;
    int end = 0;

    global_error_flag = 0;
    img_cnt = 0;
    if( argc > 1 ) {
        if( (fpin = fopen( argv[1], "rb" )) != NULL ) {
            chk_signature();
            val = get_screen_descriptor();
            if( val )
                fseek( fpin, val, SEEK_CUR ); /* skip color table */

            while( !feof( fpin )  && !global_error_flag  && !end ) {
                block_code = getbyte();
                switch( block_code ) {
                    case 0x21:  /* Extension block */
                        get_extensions();
                        break;

                    case 0x2C:  /* Image Block */
                        get_image();
                        break;
                
                    case 0x3B:  /* GIF terminator */
                        end = 1;
                        printf( "\nEnd at : %lX\n", ftell( fpin ) );
                        break;
                
                    default:
                        global_error_flag = -1;
                        break;
                }
            }
            fclose( fpin );
        }
    }
}
